//===========================================================================
// GNU C++ for LPC1114 + base board 
// show rotary encoder switches on 7-segment display
// Wouter van Ooijen - wouter@voti.nl
// license: zlib
//===========================================================================

#include "LPC11xx.h"

void wait( unsigned int t ){
   while( t > 1000 ){
      wait( 1000 ); 
      t -= 1000;
   }  
   SysTick->CTRL = 0;      // stop the timer
   SysTick->LOAD = t * 6;  // CPU clock / ( 2 * 6 ) => 1 MHz
   SysTick->VAL = 0;       // clear the timer
   SysTick->CTRL = 1;      // start the timer
   while( ( SysTick->CTRL & ( 1 << 16 ) ) == 0){} // wait for timeout

}

class output_pin {
public:
   virtual void set( bool x )=0;
   virtual void direction_set_output(){}
};

class input_pin {
public:
   virtual bool get()=0;
   virtual void direction_set_input(){}
};

enum direction { input, output };

class input_output_pin : public output_pin, public input_pin {
public:
   virtual void direction_set( direction d ){
      if( d == input ){
         direction_set_input();
      } else {
         direction_set_output();
      }
   }
   void direction_set_input(){
      direction_set( input );
   }
   void direction_set_output(){
      direction_set( output );  
   }                      
};

class hc595;

class output_pin_595 : public output_pin {
private:
   hc595 &chip;
   unsigned int pin;  
public:     
   output_pin_595( hc595 &chip, unsigned int pin ): 
      chip( chip ), pin( pin ){}    
   void set( bool x );
};


class hc595 {
private:

   output_pin &hclk, &sclk, &data;    
   unsigned int buffer;   
   
   void flush(){
      hclk.direction_set_output();  
      sclk.direction_set_output();  
      data.direction_set_output();  
      unsigned int d = buffer;
      for( unsigned int i = 0; i < 8; i++ ){
         data.set( d & 0x80 );
         sclk.set( 0 );
         wait( 1 );
         sclk.set( 1 );
         wait( 1 );
         d = d << 1;
      }   
      hclk.set( 0 );
      wait( 1 );
      hclk.set( 1 );
      wait( 1 );                             
   }
   
public:   
       
   hc595( output_pin &hclk, output_pin &sclk, output_pin &data ):
      hclk( hclk ), sclk( sclk ), data( data ), buffer( 0 ){}       
   
   void pin_set( unsigned int pin, bool x ){
      if( x ){
         buffer |= 1 << pin;   ;  
      } else {
         buffer &= ~ ( 1 << pin );  
      } 
      flush();
   }
   
   void port_set( unsigned int x ){
      buffer = x;
      flush();
   }        
   
   output_pin_595 pin( unsigned int n ){
      return output_pin_595( *this, n );    
   }
   
};

void output_pin_595::set( bool x ){
   chip.pin_set( pin, x );  
}


class input_output_pin_lpc1: public input_output_pin {
private:
   volatile int *data_pointer;
   volatile int *direction_pointer;
   unsigned int mask;
   
   static volatile int *gpioreg( unsigned int port, unsigned int offset = 0 ){
      return (volatile int *)( 0x50000000 + port * 0x10000 + offset );
   }  
   
public: 
   input_output_pin_lpc1( unsigned int port, unsigned int pin ):
      data_pointer( gpioreg( port, 0x04 << pin )),
      direction_pointer( gpioreg( port, 0x8000 )),
      mask( 1 << pin ){}
   
   void direction_set_input(){
      *direction_pointer &= ~ mask;
   }
   
   void direction_set_output(){
      *direction_pointer |= mask;  
   }
   
   void set( bool x ){
      *data_pointer = x ? -1 : 0;        
   }
   
   bool get(){ 
      return *data_pointer != 0;
   }
};

class lpc1114 {
public:
   input_output_pin_lpc1 pin( unsigned int port, unsigned int pin ){
      return input_output_pin_lpc1( port, pin );
   }
};  

void blink( output_pin &led, unsigned int delay = 200 * 1000 ){
    led.direction_set_output();
    for(;;){
       led.set( 1 );
       wait(  delay );
       led.set( 0 );
       wait(  delay );
   }      
};

         
int main( void ){   
   lpc1114 target;
   
   input_output_pin_lpc1  hclk  = target.pin( 1, 11 );
   input_output_pin_lpc1  sclk  = target.pin( 2, 11 );
   input_output_pin_lpc1  data  = target.pin( 0,  9 );
   
   hc595 sr( hclk, sclk, data ); 
   sr.port_set( 0xFF );
     
   output_pin_595 led_a = sr.pin( 1 );
   output_pin_595 led_b = sr.pin( 7 );
   led_a.direction_set_output();
   led_b.direction_set_output();
      
   input_output_pin_lpc1  rotary_a  = target.pin( 1, 0 );
   input_output_pin_lpc1  rotary_b  = target.pin( 1, 1 );
   
   rotary_a.direction_set_input();
   rotary_b.direction_set_input();
   
   for(;;){
      wait( 1000 );
      led_a.set( rotary_a.get() );
      led_b.set( rotary_b.get() );
   }
   return 0;
}
